/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 8; tab-width: 8 -*- */
/* iriver-method.c: Gnome-VFS Iriver method

   Copyright (C) 2003 Michele Baldessari 
	 
   The Gnome Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The Gnome Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the Gnome Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.

	 Authors: Michele Baldessari <michele@pupazzo.org> 
*/

#include <config.h>
#include <usb.h>
#include <string.h>

#include <libgnomevfs/gnome-vfs.h>
#include <libgnomevfs/gnome-vfs-mime.h>

#include <libgnomevfs/gnome-vfs-method.h>
#include <libgnomevfs/gnome-vfs-module.h>
#include <libgnomevfs/gnome-vfs-module-callback-module-api.h>
#include <libgnomevfs/gnome-vfs-standard-callbacks.h>

#include "iriver-i18n.h"
#include "ifp_routines.h"

#define DEBUG_IRIVER_ENABLE
#define DEBUG_IRIVER_LOCKS

#ifdef DEBUG_IRIVER_ENABLE
#define DEBUG_IRIVER(x) g_print x
#else
#define DEBUG_IRIVER(x) 
#endif

#ifdef DEBUG_IRIVER_LOCKS
#define LOCK_IRIVER() 	{g_mutex_lock (iriver_lock); g_print ("LOCK %s\n", G_GNUC_PRETTY_FUNCTION);}
#define UNLOCK_IRIVER() 	{g_print ("UNLOCK %s\n", G_GNUC_PRETTY_FUNCTION); g_mutex_unlock (iriver_lock);}
#else
#define LOCK_IRIVER() 	g_mutex_lock (iriver_lock)
#define UNLOCK_IRIVER() 	g_mutex_unlock (iriver_lock)
#endif

static GMutex *iriver_lock = NULL;
static const char *gnome_vfs_uri_get_basename (const GnomeVFSURI *uri);

#define IFP_REQ_TYPE			        0xc0
#define IFP_GET_CAPACITY_REQUEST	0x14
#define IFP_GET_FREE_REQUEST		  0x15
#define IFP_SEND_DATA							0x18
#define IFP_LS										0x0f
#define IFP_LS_NEXT								0x10
#define IFP_BULK_TO								0x03
#define IFP_BULK_FROM							0x83
#define IFP_LS_LEAVE							0x11
#define IFP_GET_FILE_INFO_SIZE		0x0b
#define IFP_FILE_INFO							0x05
#define IFP_FILE_LEAVE						0x0d
#define IFP_FILE_UPLOAD						0x06
#define IFP_FILE_DOWNLOAD					0x07
#define IFP_FILE_UPLOAD_APPEND		0x08
#define IFP_FILE_DELETE						0x0e
#define IFP_DIR_DELETE						0x13
#define IFP_DIR_CREATE						0x12
#define IFP_FORMAT								0x16
#define IFP_FIRMWARE_UPDATE				0x17
#define IFP_02_COMMAND						0x02
#define IFP_02_BATTERY						0x08
#define IFP_02_STRING							0x00
#define IFP_02_FIRMWARE						0x03

#define TOUT 1000

#define IFP_BULK_MAXPATHLEN				0x400
#define IFP_MAXPATHLEN						0x200



#define NUM_RESULT_STRINGS 41
static const char * const
result_strings[NUM_RESULT_STRINGS] = {
	"GNOME_VFS_OK",
	"GNOME_VFS_ERROR_NOT_FOUND",
	"GNOME_VFS_ERROR_GENERIC",
	"GNOME_VFS_ERROR_INTERNAL",
	"GNOME_VFS_ERROR_BAD_PARAMETERS",
	"GNOME_VFS_ERROR_NOT_SUPPORTED",
	"GNOME_VFS_ERROR_IO",
	"GNOME_VFS_ERROR_CORRUPTED_DATA",
	"GNOME_VFS_ERROR_WRONG_FORMAT",
	"GNOME_VFS_ERROR_BAD_FILE",
	"GNOME_VFS_ERROR_TOO_BIG",
	"GNOME_VFS_ERROR_NO_SPACE",
	"GNOME_VFS_ERROR_READ_ONLY",
	"GNOME_VFS_ERROR_INVALID_URI",
	"GNOME_VFS_ERROR_NOT_OPEN",
	"GNOME_VFS_ERROR_INVALID_OPEN_MODE",
	"GNOME_VFS_ERROR_ACCESS_DENIED",
	"GNOME_VFS_ERROR_TOO_MANY_OPEN_FILES",
	"GNOME_VFS_ERROR_EOF",
	"GNOME_VFS_ERROR_NOT_A_DIRECTORY",
	"GNOME_VFS_ERROR_IN_PROGRESS",
	"GNOME_VFS_ERROR_INTERRUPTED",
	"GNOME_VFS_ERROR_FILE_EXISTS",
	"GNOME_VFS_ERROR_LOOP",
	"GNOME_VFS_ERROR_NOT_PERMITTED",
	"GNOME_VFS_ERROR_IS_DIRECTORY",
	"GNOME_VFS_ERROR_NO_MEMORY",
	"GNOME_VFS_ERROR_HOST_NOT_FOUND",
	"GNOME_VFS_ERROR_INVALID_HOST_NAME",
	"GNOME_VFS_ERROR_HOST_HAS_NO_ADDRESS",
	"GNOME_VFS_ERROR_LOGIN_FAILED",
	"GNOME_VFS_ERROR_CANCELLED",
	"GNOME_VFS_ERROR_DIRECTORY_BUSY",
	"GNOME_VFS_ERROR_DIRECTORY_NOT_EMPTY",
	"GNOME_VFS_ERROR_TOO_MANY_LINKS",
	"GNOME_VFS_ERROR_READ_ONLY_FILE_SYSTEM",
	"GNOME_VFS_ERROR_NOT_SAME_FILE_SYSTEM",
	"GNOME_VFS_ERROR_NAME_TOO_LONG",
	"GNOME_VFS_ERROR_SERVICE_NOT_AVAILABLE",
	"GNOME_VFS_NUM_ERRORS"
};

static struct stat status;
static struct usb_bus *bus;
static struct usb_device *tmp_dev;
static struct usb_device *dev = NULL;
static usb_dev_handle *dh = NULL;
static gboolean found_iriver = FALSE;

typedef struct {
	GList *list;
	GList *mime;
	GList *size;
	GList *type;
	gchar *dirname;
	guint pos;
	gboolean list_type_is_folder;
} DirectoryHandle;

typedef struct {
	gboolean is_read;
	/* handle for the temp file */
	GnomeVFSHandle *handle;
	/* For writing */
	gchar *filename;
	gint folder;
} FileHandle;

/* Module entry points. */
GnomeVFSMethod *vfs_module_init     (const char     *method_name,
					   										     const char     *args);
void            vfs_module_shutdown (GnomeVFSMethod *method);

const char *
gnome_vfs_uri_get_basename (const GnomeVFSURI *uri)
{
	char *p;

	g_return_val_if_fail (uri != NULL, NULL);

	if (uri->text == NULL) {
		return NULL;
	}

	p = strchr (gnome_vfs_uri_get_path(uri), GNOME_VFS_URI_PATH_CHR);
	if (p == NULL) {
		return NULL;
	}

	p++;
	if (*p == '\0') {
		return NULL;
	}

	return p;
}

static char *
unix_filename_to_dos (const char *filename)
{
	int len, i;
	char *dos_filename;
	
	dos_filename = g_strdup (filename);
	len = strlen (dos_filename);
	for (i = 0; i < len; i++)
		if (dos_filename[i] == '/')
			dos_filename[i] = '\\';
	
	return dos_filename;
}

static GnomeVFSResult
do_get_file_info (GnomeVFSMethod *method,
		  GnomeVFSURI *uri,
		  GnomeVFSFileInfo *file_info,
		  GnomeVFSFileInfoOptions options,
		  GnomeVFSContext *context)

{
	gchar *name, *tmp;
	unsigned char ctl[4];
	unsigned char ctl8[8];
  int size = 0;
	DEBUG_IRIVER (("do_get_file_info() %s\n",
				gnome_vfs_uri_to_string (uri, 0)));
  LOCK_IRIVER();

	/* Is this root? */
	if (gnome_vfs_uri_has_parent (uri) == 0) {
		file_info->name = g_strdup ("/");
		file_info->valid_fields = 
			GNOME_VFS_FILE_INFO_FIELDS_TYPE |
			GNOME_VFS_FILE_INFO_FIELDS_MIME_TYPE;
		file_info->type = GNOME_VFS_FILE_TYPE_DIRECTORY;
		file_info->mime_type = g_strdup ("x-directory/normal");
		UNLOCK_IRIVER();
		return GNOME_VFS_OK;
	}

	name = g_strdup(gnome_vfs_uri_get_basename(uri));
		
	DEBUG_IRIVER (("do_get_file_info() base %s\n", name));
	remove_double_slash(name);
	if (g_str_has_suffix (name, GNOME_VFS_URI_PATH_STR) == TRUE) {
		g_free(name);
		name = g_strdup(gnome_vfs_uri_extract_dirname(uri));
	}

  name = unix_filename_to_dos(name);
	DEBUG_IRIVER (("do_get_file_info() base %s\n", name));
	/* Check if file exists */
	if (ifp_send_filename(dh, name) < 0) {
		g_free(name);
		UNLOCK_IRIVER();
		return GNOME_VFS_ERROR_NOT_FOUND;
	}
	usb_control_msg(dh,IFP_REQ_TYPE,IFP_LS,0,0,ctl,sizeof(ctl),TOUT);
	if (!ifp_ok(ctl)) { /* a file maybe then */
		usb_control_msg(dh,IFP_REQ_TYPE,IFP_FILE_INFO,1,0,ctl,sizeof(ctl),TOUT);
		if (!ifp_ok(ctl)) {
			g_free(name);
			UNLOCK_IRIVER();
			return GNOME_VFS_ERROR_NOT_FOUND;
		}
		usb_control_msg(dh,IFP_REQ_TYPE,IFP_GET_FILE_INFO_SIZE,0,0,ctl8,sizeof(ctl8),TOUT);
		size = ctl8[4] + (ctl8[5] << 8) + (ctl8[6] << 16) + (ctl8[7] << 24);
	
		file_info->name = gnome_vfs_uri_extract_short_name (uri);
		file_info->valid_fields = 
			GNOME_VFS_FILE_INFO_FIELDS_TYPE |
			GNOME_VFS_FILE_INFO_FIELDS_MIME_TYPE |
			GNOME_VFS_FILE_INFO_FIELDS_SIZE;
		file_info->type = GNOME_VFS_FILE_TYPE_REGULAR;
		tmp = g_ascii_strdown(file_info->name, -1);
		if (g_str_has_suffix(tmp, "mp3") == TRUE)
			file_info->mime_type = g_strdup("audio/x-mp3");
		else if (g_str_has_suffix(tmp, "ogg") == TRUE) // Wishful thinking *g*
			file_info->mime_type = g_strdup("audio/x-ogg");
		else if (g_str_has_suffix(tmp, "wma") == TRUE)
			file_info->mime_type = g_strdup("audio/x-ms-wma");
		else if (g_str_has_suffix(tmp, "asf") == TRUE)
			file_info->mime_type = g_strdup("audio/x-ms-asf");
		else 
			file_info->mime_type = g_strdup("application/octet-stream");

		g_free(tmp);
		file_info->size = size;
		usb_control_msg(dh,IFP_REQ_TYPE,IFP_FILE_LEAVE,0,0,ctl,sizeof(ctl),TOUT);
		if (!ifp_ok(ctl)) {
			g_free (name);
			UNLOCK_IRIVER();
			return GNOME_VFS_ERROR_GENERIC;
		}
	} else { /* A directory ... */
		file_info->name = gnome_vfs_uri_extract_short_name (uri);
		file_info->type = GNOME_VFS_FILE_TYPE_DIRECTORY;
		file_info->valid_fields = 
			GNOME_VFS_FILE_INFO_FIELDS_TYPE |
			GNOME_VFS_FILE_INFO_FIELDS_MIME_TYPE |
			GNOME_VFS_FILE_INFO_FIELDS_SIZE;
		file_info->mime_type = g_strdup ("x-directory/normal");
		file_info->size = 0;
	}
	g_free(name);
	usb_control_msg(dh,IFP_REQ_TYPE,IFP_LS_LEAVE,0,0,ctl,sizeof(ctl),TOUT);
	UNLOCK_IRIVER();
	return GNOME_VFS_OK;
}

static GnomeVFSResult
do_open (GnomeVFSMethod *method,
	 GnomeVFSMethodHandle **method_handle,
	 GnomeVFSURI *uri,
	 GnomeVFSOpenMode mode,
	 GnomeVFSContext *context)
{
	FileHandle *handle = NULL;
	GnomeVFSResult res;
	GnomeVFSFileInfo *info;
	gint card, folder, song;
	gboolean is_read = FALSE;
	gchar  *name, *local;

	DEBUG_IRIVER(("do_open() %s mode %d\n",
				gnome_vfs_uri_to_string (uri, 0), mode));

	local = g_strdup_printf ("%s/iriver_tmp_file.mp3", g_get_tmp_dir());
  if (g_file_test (local, G_FILE_TEST_EXISTS) == TRUE)
		return GNOME_VFS_ERROR_TOO_MANY_OPEN_FILES;

	if (mode & GNOME_VFS_OPEN_READ)
		is_read = TRUE;
	else if (mode & GNOME_VFS_OPEN_WRITE)
		is_read = FALSE;
	else
		return GNOME_VFS_ERROR_INVALID_OPEN_MODE;
	
	if (is_read == TRUE) 
		DEBUG_IRIVER (("Is read-only open...\n"));

	info = gnome_vfs_file_info_new ();
	res = do_get_file_info(method, uri, info, 0, context);
	LOCK_IRIVER();
	if ((res != GNOME_VFS_OK) && (is_read == TRUE)) {
		UNLOCK_IRIVER();
		g_free (info);
		g_free (local);
		DEBUG_IRIVER (("do_open(): file not found on read open, result\n"));
		return res;
	}
	
	if ((res == GNOME_VFS_ERROR_NOT_FOUND) && (is_read == TRUE)) {
		UNLOCK_IRIVER();
		g_free (info);
		g_free (local);
		DEBUG_IRIVER (("do_open(): file not found on read open\n"));
		return GNOME_VFS_ERROR_NOT_FOUND;
	}

	if ((info->type == GNOME_VFS_FILE_TYPE_DIRECTORY) && (is_read == TRUE)) {
		UNLOCK_IRIVER();
		g_free (info);
		g_free (local);
		DEBUG_IRIVER(("do_open(): file is a directory on a read open\n"));
		return GNOME_VFS_ERROR_IS_DIRECTORY;
	}
	
	if (is_read == TRUE) {
		GnomeVFSHandle *local_handle;
		gint res, i;
		FILE *f;
		
		name = g_strdup(gnome_vfs_uri_get_path(uri));
		remove_double_slash(name);
		name = unix_filename_to_dos(name);
		
		DEBUG_IRIVER(("do_open(): opening read-only\n"));
		/* Copying file to /tmp */	
		f = fopen(local, "w");
		if (f == NULL) {
			g_free (name);
			UNLOCK_IRIVER ();
			return GNOME_VFS_ERROR_IO;
		}
		i = ifp_download_file(dh, name, f);
		if (i < 0) {
			g_free (name);
			UNLOCK_IRIVER ();
			return GNOME_VFS_ERROR_IO;
		}
		fclose(f);	
		DEBUG_IRIVER (("do_open(): copied to tmp\n"));

		if (gnome_vfs_open (&local_handle, local, GNOME_VFS_OPEN_READ) !=
			GNOME_VFS_OK) {
			unlink(local);
			g_free (name);
			UNLOCK_IRIVER();
			DEBUG_IRIVER (("do_open(): error opening tmp file\n"));
			return GNOME_VFS_ERROR_IO;
		}
		
		DEBUG_IRIVER (("do_open(): opened local file\n"));

		handle = g_new(FileHandle, 1);
		handle->is_read = is_read;
		handle->handle = local_handle;
		handle->filename = NULL;
		g_free (name);
	} else { /* Open for writing */
		GnomeVFSHandle *local_handle;
		gchar *basename, *local;

		DEBUG_IRIVER(("do_open: opening write\n"));

		basename = g_strdup (gnome_vfs_uri_get_basename (uri));
    local = g_strdup_printf ("file://%s/%s", g_get_tmp_dir(), "iriver_tmp_file.mp3");
	
		DEBUG_IRIVER (("open_write: %s\n", local));
		if (gnome_vfs_create (&local_handle, local,
				GNOME_VFS_OPEN_WRITE, FALSE, 0600) != GNOME_VFS_OK) {
			UNLOCK_IRIVER();
			DEBUG_IRIVER (("do_open(): error opening the file for writing %s\n",
					local));
			g_free(local);
			g_free(basename);
			return GNOME_VFS_ERROR_IO;
		}

		/* Create the handle */
		handle = g_new(FileHandle, 1);
		handle->is_read = is_read;
		handle->handle = local_handle;
		handle->filename = g_strdup(basename);
		g_free(local);
		g_free(basename);
		
	}
	*method_handle = (GnomeVFSMethodHandle *)handle;
	
	return GNOME_VFS_OK;
}

static GnomeVFSResult
do_close (GnomeVFSMethod *method,
	  GnomeVFSMethodHandle *method_handle,
	  GnomeVFSContext *context)
{
	GnomeVFSURI *uri;
	FileHandle *handle = (FileHandle *)method_handle;
	GnomeVFSResult res = GNOME_VFS_OK;
  gchar *tmpstr, *dest;

	tmpstr = g_strdup_printf("%s/iriver_tmp_file.mp3", g_get_tmp_dir());
	if (handle->is_read == TRUE) {
		gint tmp;
		DEBUG_IRIVER(("do_close() read\n"));

		tmp = unlink (tmpstr);
		DEBUG_IRIVER (("do_close(): unlinked temp file %s - %d \n", 
				tmpstr, tmp));

		if (tmp != 0)
			res = GNOME_VFS_ERROR_IO;
		else
			res = GNOME_VFS_OK;
	} else { /* Ending a write */
		gint tmp, i;
		FILE *f;
		struct stat stat;
		
		DEBUG_IRIVER(("do_close() write \n"));
/*		uri =	gnome_vfs_uri_new (handle->filename);
		dest = gnome_vfs_uri_to_string (g_strdup(gnome_vfs_uri_get_basename(uri)); */

		dest = g_strdup_printf("/%s", handle->filename);
		DEBUG_IRIVER (("do_close(): uploading song %s to %s .. %s \n",
					tmpstr, handle->filename, dest));
		f = fopen(tmpstr, "r");
		if (f == NULL) {
			UNLOCK_IRIVER();
			g_free(tmpstr);
			return GNOME_VFS_ERROR_GENERIC;
		}
		DEBUG_IRIVER (("fileno %d\n", fileno(f)));
		fstat(fileno(f), &stat);
		DEBUG_IRIVER (("fileno %d- size %d\n", fileno(f), stat.st_size));
		
		i = ifp_upload_file(dh, dest, f);
		fclose(f);
		if (i <= -2) {
			UNLOCK_IRIVER();
			g_free(tmpstr);
			return GNOME_VFS_ERROR_NO_SPACE;
		}
		else if (i<0) {
			UNLOCK_IRIVER();
			g_free(tmpstr);
			return GNOME_VFS_ERROR_IO;
		}

		unlink (tmpstr);
		DEBUG_IRIVER (("do_close(): uploading result: %d\n", res));
	}

	gnome_vfs_close (handle->handle);
	g_free (handle->filename);
	g_free (handle);

	UNLOCK_IRIVER();
	g_free(tmpstr);

	return res;
}

static GnomeVFSResult
do_read (GnomeVFSMethod *method,
	 GnomeVFSMethodHandle *method_handle,
	 gpointer buffer,
	 GnomeVFSFileSize num_bytes,
	 GnomeVFSFileSize *bytes_read,
	 GnomeVFSContext *context)
{
	FileHandle *handle = (FileHandle *)method_handle;

	DEBUG_IRIVER (("do_read() %Lu bytes\n", num_bytes));

	return gnome_vfs_read (handle->handle, buffer, num_bytes, bytes_read);
}

static GnomeVFSResult
do_write (GnomeVFSMethod *method,
	  GnomeVFSMethodHandle *method_handle,
	  gconstpointer buffer,
	  GnomeVFSFileSize num_bytes,
	  GnomeVFSFileSize *bytes_written,
	  GnomeVFSContext *context)
{
	FileHandle *handle = (FileHandle *)method_handle;

	DEBUG_IRIVER (("1- do_write() %p - handle %d - name %s - num_bytes %d - bytes_written %d\n", method_handle, handle, handle->filename, num_bytes, bytes_written));

	return gnome_vfs_write (handle->handle, buffer, num_bytes,
			bytes_written);
}

static GnomeVFSResult
do_create (GnomeVFSMethod *method,
	   GnomeVFSMethodHandle **method_handle,
	   GnomeVFSURI *uri,
	   GnomeVFSOpenMode mode,
	   gboolean exclusive,
	   guint perm,
	   GnomeVFSContext *context)

{
	DEBUG_IRIVER (("do_create() %s mode %d\n",
				gnome_vfs_uri_to_string (uri, 0), mode));


	return do_open (method, method_handle, uri, GNOME_VFS_OPEN_WRITE,
			context);
}



static GnomeVFSResult
do_open_directory (GnomeVFSMethod *method,
		   GnomeVFSMethodHandle **method_handle,
		   GnomeVFSURI *uri,
		   GnomeVFSFileInfoOptions options,
		   GnomeVFSContext *context)
{
	DirectoryHandle *directory_handle;
	GList *list = NULL;
	GList *size = NULL;
	GList *mime = NULL;
	GList *type = NULL;
  unsigned char ctl[4];
  unsigned char ctl8[8];
  unsigned char inbuf[0x100];
  unsigned char strbuf[0x200];
  unsigned char strbuf2[0x200];
  unsigned char strnew[0x200];
  char *dir;
	int i;
  int ret;
  int s;
	char *name, *tmp;
	unsigned char prefix[0x100];

	DEBUG_IRIVER(("do_open_directory() %s\n",
		gnome_vfs_uri_to_string (uri, 0)));
	
	name = g_strdup(gnome_vfs_uri_get_basename(uri));
	if (name == NULL) {
		name = (char *)malloc(3);
		strcpy(name, "/");
	}
  DEBUG_IRIVER(("do_open_directory() base %s\n",
		name));
  LOCK_IRIVER();

	/* Directory exists? */
	if (ifp_send_filename(dh, name) < 0) {
		g_free (name);
		return GNOME_VFS_ERROR_NOT_FOUND;
	}
  usb_control_msg(dh,IFP_REQ_TYPE,IFP_LS,0,0,ctl,sizeof(ctl),TOUT);
	if (!ifp_ok(ctl)) {
		g_free(name);
		return GNOME_VFS_ERROR_IO;
	}
	
	while (1) {
  	usb_control_msg(dh,IFP_REQ_TYPE,IFP_LS_NEXT,3,0,ctl,sizeof(ctl),TOUT);	// 3=magic !!!???
	  ret = ifp_ctl_ret(ctl);
	  switch (ret) {
	    case 0:
		    goto LS_END;
	    case 1: /* File */
		    usb_bulk_read(dh,IFP_BULK_FROM,inbuf,sizeof(inbuf),TOUT);
		    unicode2string(strbuf,inbuf);
				strcpy(strbuf2,name);
				if (strcmp(name,"/") != 0) strcat(strbuf2,"/");
				strcat(strbuf2,strbuf); // path/file
				ifp_send_filename(dh,strbuf2);
				usb_control_msg(dh,IFP_REQ_TYPE,IFP_FILE_INFO,1,0,ctl,sizeof(ctl),TOUT);
				if (!ifp_ok(ctl)) {
					return GNOME_VFS_ERROR_IO;
				}
				usb_control_msg(dh,IFP_REQ_TYPE,IFP_GET_FILE_INFO_SIZE,0,0,ctl8,sizeof(ctl8),TOUT);
				usb_control_msg(dh,IFP_REQ_TYPE,IFP_FILE_LEAVE,0,0,ctl,sizeof(ctl),TOUT); //??????

				s = ctl8[4] + (ctl8[5] << 8) + (ctl8[6] << 16) + (ctl8[7] << 24);
				list = g_list_append(list, g_strdup(strbuf));
				/* We base our mime detecting stuff just on extension.. Broken? */
				tmp = g_ascii_strdown(strbuf, -1);
				if (g_str_has_suffix(tmp, "mp3") == TRUE)
					mime = g_list_append(mime, g_strdup("audio/x-mp3"));
				else if (g_str_has_suffix(tmp, "ogg") == TRUE) // Wishful thinking *g*
					mime = g_list_append(mime, g_strdup("audio/x-ogg"));
				else if (g_str_has_suffix(tmp, "wma") == TRUE)
					mime = g_list_append(mime, g_strdup("audio/x-ms-wma"));
				else if (g_str_has_suffix(tmp, "asf") == TRUE)
					mime = g_list_append(mime, g_strdup("audio/x-ms-asf"));
				else 
					mime = g_list_append(mime, g_strdup("application/octet-stream"));
				g_free(tmp);
				size = g_list_append(size, (char *)s);
				type = g_list_append(type, (char *)GNOME_VFS_FILE_TYPE_REGULAR);
				break;
	    case 2: /* Directory */
				usb_bulk_read(dh,IFP_BULK_FROM,inbuf,sizeof(inbuf),TOUT);
				unicode2string(strbuf,inbuf);
				list = g_list_append(list, g_strdup(strbuf));
				mime = g_list_append(mime, g_strdup("x-directory/normal"));
				size = g_list_append(size, (char *)0);
				type = g_list_append(type, (char *)GNOME_VFS_FILE_TYPE_DIRECTORY);
				break;
	    default:
				return GNOME_VFS_ERROR_SERVICE_NOT_AVAILABLE;
		}
  }
LS_END:
  usb_control_msg(dh,IFP_REQ_TYPE,IFP_LS_LEAVE,0,0,ctl,sizeof(ctl),TOUT);

	directory_handle = g_new0 (DirectoryHandle, 1);
	directory_handle->dirname = g_strdup(name);
	directory_handle->pos = 0;
	directory_handle->list = list;
	directory_handle->mime = mime;
	directory_handle->size = size;
	directory_handle->type = type;

	*method_handle = (GnomeVFSMethodHandle *) directory_handle;

	g_free(name);
	UNLOCK_IRIVER();
	DEBUG_IRIVER(("do_open_directory end\n"));
	return GNOME_VFS_OK;
}


static GnomeVFSResult
do_close_directory (GnomeVFSMethod *method,
		    GnomeVFSMethodHandle *method_handle,
		    GnomeVFSContext *context)
{
	DirectoryHandle *directory_handle = (DirectoryHandle *) method_handle;
	DEBUG_IRIVER(("do_close_directory \n"));
	if (directory_handle = NULL)
		return GNOME_VFS_OK;
	if (directory_handle->dirname != NULL)
		g_free (directory_handle->dirname);
	/* FIXME: need to free correctly all three lists */
	if (directory_handle != NULL)
		g_free (directory_handle);

	DEBUG_IRIVER(("do_close_directory: end\n"));
	return GNOME_VFS_OK;
}


static GnomeVFSResult
do_read_directory (GnomeVFSMethod *method,
		   GnomeVFSMethodHandle *method_handle,
		   GnomeVFSFileInfo *file_info,
		   GnomeVFSContext *context)
{
	DirectoryHandle *directory_handle = (DirectoryHandle *) method_handle;
	GList *list, *mime, *size, *type;

	DEBUG_IRIVER (("do_read_directory()\n"));
	LOCK_IRIVER();
	/*DEBUG_IRIVER (("Pos %d - list len %d \n", directory_handle->pos, 
			g_list_length(directory_handle->list)));*/
	if (directory_handle->pos > g_list_length(directory_handle->list)) {
		UNLOCK_IRIVER();
		return GNOME_VFS_ERROR_EOF;
	}
  list = g_list_nth(directory_handle->list, directory_handle->pos);
  mime = g_list_nth(directory_handle->mime, directory_handle->pos);
  size = g_list_nth(directory_handle->size, directory_handle->pos);
	type = g_list_nth(directory_handle->type, directory_handle->pos);
	if (list == NULL) {
		UNLOCK_IRIVER();
		return GNOME_VFS_ERROR_EOF;
	}

  GNOME_VFS_FILE_INFO_SET_LOCAL(file_info, FALSE);
	
	file_info->valid_fields = GNOME_VFS_FILE_INFO_FIELDS_NONE |
		GNOME_VFS_FILE_INFO_FIELDS_TYPE |
		GNOME_VFS_FILE_INFO_FIELDS_MIME_TYPE;

  if ((int)type->data == GNOME_VFS_FILE_TYPE_DIRECTORY) {
		file_info->type = GNOME_VFS_FILE_TYPE_DIRECTORY;
	} else {
		file_info->type = GNOME_VFS_FILE_TYPE_REGULAR;
		file_info->valid_fields |= GNOME_VFS_FILE_INFO_FIELDS_SIZE;
		file_info->size = (int)(size->data);
		DEBUG_IRIVER (("Size %d\n", (int)size->data));
	}

	file_info->name = (gchar *)g_strdup(list->data);
	file_info->mime_type = g_strdup(mime->data);

	directory_handle->pos++;
	UNLOCK_IRIVER();
	return GNOME_VFS_OK;
}


static gboolean
do_is_local (GnomeVFSMethod *method,
	     const GnomeVFSURI *uri)
{
	return FALSE;
}


static GnomeVFSResult
do_unlink (GnomeVFSMethod *method,
			GnomeVFSURI *uri,
			GnomeVFSContext *context)
{
	char *name;
	unsigned char ctl[4];
	GnomeVFSResult res;
	GnomeVFSFileInfo *info;

	DEBUG_IRIVER (("do_unlink() \n"));
	
	/* FIXME: this needs to take into account if it's an empty
	 * dir or not..*/
	info = gnome_vfs_file_info_new ();
	res = do_get_file_info(method, uri, info, 0, context);

	if (res == GNOME_VFS_ERROR_NOT_FOUND) 
		return GNOME_VFS_ERROR_NOT_FOUND;
	
	
	LOCK_IRIVER();
	name = g_strdup(gnome_vfs_uri_get_basename(uri));
	DEBUG_IRIVER (("do_unlink() starting removal of %s\n", name));
	if (name == NULL) {
		name = (char *)malloc(3);
		strcpy(name, "/");
	}
	if (ifp_send_filename(dh, name) < 0) {
		UNLOCK_IRIVER();
		g_free(name);
		return GNOME_VFS_ERROR_NOT_FOUND;
	}

	if ((info->type == GNOME_VFS_FILE_TYPE_DIRECTORY)) {
		usb_control_msg(dh, IFP_REQ_TYPE, IFP_DIR_DELETE, 0, 0, ctl, sizeof(ctl), TOUT);
		if (!ifp_ok(ctl)) {
			UNLOCK_IRIVER();
			g_free(name);
			return GNOME_VFS_ERROR_IO;	
		}
	} else {
	  usb_control_msg(dh, IFP_REQ_TYPE, IFP_FILE_DELETE, 0, 0, ctl, sizeof(ctl), TOUT);
		if (!ifp_ok(ctl)) {
			UNLOCK_IRIVER();
			g_free(name);
			return GNOME_VFS_ERROR_IO;
		}
	}

	UNLOCK_IRIVER();
	g_free(name);
	return GNOME_VFS_OK;
}

static GnomeVFSResult
do_check_same_fs (GnomeVFSMethod *method,
		  GnomeVFSURI *a,
		  GnomeVFSURI *b,
		  gboolean *same_fs_return,
		  GnomeVFSContext *context)
{
	DEBUG_IRIVER (("do_check_same_fs()\n"));
	*same_fs_return = FALSE;

	return GNOME_VFS_OK;
}

static GnomeVFSMethod method = {
	sizeof (GnomeVFSMethod),
       do_open,              /* do_open */
       do_create,            /* do_create */
       do_close,             /* do_close */
       do_read,              /* do_read  */
       do_write,             /* do_write */
       NULL,                 /* do_seek */
       NULL,                 /* do_tell */
       NULL,                 /* do_truncate_handle */
       do_open_directory,    /* do_open_directory */
       do_close_directory,   /* do_close_directory */
       do_read_directory,    /* do_read_directory */
       do_get_file_info,     /* do_get_file_info */
       NULL,                 /* do_get_file_info_from_handle */
       do_is_local,          /* do_is_local */
       NULL,                 /* do_make_directory */
       NULL,                 /* do_remove_directory */
       NULL,                 /* do_move */
       do_unlink,            /* do_unlink */
       do_check_same_fs,     /* do_check_same_fs */
       NULL,                 /* do_set_file_info */
       NULL,                 /* do_truncate */
       NULL,                 /* do_find_directory */
       NULL                  /* do_create_symbolic_link */
};

GnomeVFSMethod *
vfs_module_init (const char *method_name, const char *args)
{
	/* Set debug level */
  usb_init();
  usb_find_busses();
  if (usb_find_devices() <= 0) {
    DEBUG_IRIVER (("No usb devices found.\n"));
    return NULL;
  }

  for (bus = usb_busses; bus; bus = bus->next) {
    for (tmp_dev = bus->devices; tmp_dev; tmp_dev = tmp_dev->next) {
      if ((tmp_dev->descriptor.idVendor == iRiver_Vendor) &&
        (tmp_dev->descriptor.idProduct == iFP_Product)) {
        dev = tmp_dev;
        goto IFPFOUND;
      }
    }
  }
	
IFPFOUND:
  if (dev) {
    dh = usb_open(dev);
		DEBUG_IRIVER (("Battery %d \n", ifp_battery_condition(dh)));
    if ((usb_claim_interface(dh, 0) == 0) && ifp_check_connect(dh)) {
			DEBUG_IRIVER (("Capacity %d bytes - Free %d bytes\n", ifp_get_capacity(dh),
			ifp_get_free(dh)));
			iriver_lock = g_mutex_new();
			found_iriver = TRUE;
    } else {
      if (dh) usb_close(dh);
      fprintf(stderr, "Unable to connect to IFP");
      return NULL;
    }
	} else 
		return NULL;
	return &method;
}

void
vfs_module_shutdown (GnomeVFSMethod *method)
{
	DEBUG_IRIVER (("Module unloaded!\n"));
	g_mutex_free(iriver_lock);
	if (found_iriver) {
		usb_close(dh);
	}
}
