#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fts.h>
#include <unistd.h>

#include "unicodehack.h"
#include "ifp_routines.h"

#define IFP_REQ_TYPE			0xc0
#define IFP_GET_CAPACITY_REQUEST	0x14
#define IFP_GET_FREE_REQUEST		0x15
#define IFP_SEND_DATA			0x18
#define IFP_LS				0x0f
#define IFP_LS_NEXT			0x10
#define IFP_BULK_TO			0x03
#define IFP_BULK_FROM			0x83
#define IFP_LS_LEAVE			0x11
#define IFP_GET_FILE_INFO_SIZE		0x0b
#define IFP_FILE_INFO			0x05
#define IFP_FILE_LEAVE			0x0d
#define IFP_FILE_UPLOAD			0x06
#define IFP_FILE_DOWNLOAD		0x07
#define IFP_FILE_UPLOAD_APPEND		0x08
#define IFP_FILE_DELETE			0x0e
#define IFP_DIR_DELETE			0x13
#define IFP_DIR_CREATE			0x12
#define IFP_FORMAT			0x16
#define IFP_FIRMWARE_UPDATE		0x17
#define IFP_02_COMMAND			0x02
#define IFP_02_BATTERY			0x08
#define IFP_02_STRING			0x00
#define IFP_02_FIRMWARE			0x03

#define TOUT 1000

#define IFP_BULK_MAXPATHLEN		0x400
#define IFP_MAXPATHLEN			0x200


int ifp_send_data(usb_dev_handle *dh, char *data, size_t size);
int push_dir(char *name);
int pop_dir(char *name);
int comparefts(const FTSENT **a, const FTSENT **b);
void remove_double_slash(char *path);

typedef struct dirlists {
    struct dirlists *next;
    char *name;
} dirlist;

dirlist *list = NULL;

int convert_to_backslash(char *dest,const char *src) {
  int i = 0, count_slash = 0;
  while (src[i] != '\0') {
    if ((src[i] == '/') && (src[i+1] != '\0')) {
      count_slash++;
    } else if ((src[i] == '/') && (src[i+1] == '\0')) { /* String ends with / */
      *dest++ = '\\';  
    } else {
      if (count_slash > 0) 
        *dest++ = '\\';
      else
        *dest++ = src[i];
      count_slash = 0;
    }
    ++i;
  }
  *dest = '\0';
  return i;
}


int slash2backslash(char *dest,const char *src) {
    int i = 0;

    remove_double_slash((char*)src);
    while (src[i] != 0) {
	if (src[i] == '/') dest[i++] = '\\'; else dest[i] = src[i++];
    }
    dest[i] = 0;
    return i;
}

int ifp_ctl_ret(unsigned char *control) {
    return (control[0] + 256*control[1]);
}

int ifp_ok(unsigned char *control) {
    return (ifp_ctl_ret(control) == 1);
}

int ifp_get_capacity(usb_dev_handle *dh) {
    int i;
    unsigned char ctl[8];
    i = usb_control_msg(dh,IFP_REQ_TYPE,IFP_GET_CAPACITY_REQUEST,0,0,(char *)&ctl,sizeof(ctl),TOUT);
    if (!ifp_ok(ctl)) return -1;
    return ctl[7]*256*256*256 + ctl[6]*256*256 + ctl[5]*256 + ctl[4];
}

int ifp_get_free(usb_dev_handle *dh) {
    int i;
    unsigned char ctl[8];
    i = usb_control_msg(dh,IFP_REQ_TYPE,IFP_GET_FREE_REQUEST,0,0,ctl,sizeof(ctl),TOUT);
    if (!ifp_ok(ctl)) return -1;
    return ctl[7]*256*256*256 + ctl[6]*256*256 + ctl[5]*256 + ctl[4];
}

int ifp_send_filename(usb_dev_handle *dh, const char *name) {
    unsigned char buf[0x400];
    unsigned char strbuf[0x200];
    //slash2backslash(strbuf,name);
    convert_to_backslash(strbuf,name);
    string2unicode(buf,strbuf);
    return ifp_send_data(dh,buf,sizeof(buf));
}

int ifp_send_data(usb_dev_handle *dh, char *data, size_t size) {
    unsigned char ctl[8];
    usb_control_msg(dh,IFP_REQ_TYPE,IFP_SEND_DATA,size,0,ctl,sizeof(ctl),TOUT);
    if (!ifp_ok(ctl)) return -1;
    //printf("Send data - %d bytes want, %d allowed\n",size,ctl[4]+ctl[5]*256);
    usb_bulk_write(dh,IFP_BULK_TO,data,size,TOUT);
    return 0;
}

int ifp_list_dir_printf(usb_dev_handle *dh, const char *name, int format, int recurrent) {
    unsigned char ctl[4];
    unsigned char ctl8[8];
    unsigned char inbuf[0x100];
    unsigned char strbuf[0x200];
    unsigned char strbuf2[0x200];
    unsigned char strnew[0x200];
    char *dir;
    int ret;
    int size;
    const char *prefix = (strcmp(name,"/") == 0) ? (name+1) : name;

    if (ifp_send_filename(dh,name) < 0) return -1;
    
    usb_control_msg(dh,IFP_REQ_TYPE,IFP_LS,0,0,ctl,sizeof(ctl),TOUT);
    if (!ifp_ok(ctl)) return -1;
    while (1) {
	usb_control_msg(dh,IFP_REQ_TYPE,IFP_LS_NEXT,3,0,ctl,sizeof(ctl),TOUT);	// 3=magic !!!???
	ret = ifp_ctl_ret(ctl);
	switch (ret) {
	    case 0:
		goto LS_END;
	    case 1:
		usb_bulk_read(dh,IFP_BULK_FROM,inbuf,sizeof(inbuf),TOUT);
		unicode2string(strbuf,inbuf);
		strcpy(strbuf2,name);
		if (strcmp(name,"/") != 0) strcat(strbuf2,"/");
		strcat(strbuf2,strbuf); // path/file
		ifp_send_filename(dh,strbuf2);
		usb_control_msg(dh,IFP_REQ_TYPE,IFP_FILE_INFO,1,0,ctl,sizeof(ctl),TOUT);
		if (!ifp_ok(ctl))
			fprintf(stderr, "error\n");
		usb_control_msg(dh,IFP_REQ_TYPE,IFP_GET_FILE_INFO_SIZE,0,0,ctl8,sizeof(ctl8),TOUT);
		//printf("ctl8: %d %d %d %d %d %d %d %d\n",ctl8[0],ctl8[1],ctl8[2],ctl8[3],ctl8[4],ctl8[5],ctl8[6],ctl8[7],ctl8[8]);
		size = ctl8[4] + ctl8[5]*256 + ctl8[6]*256*256 + ctl8[7]*256*256*256;
		if (format)
		    printf("-rw-rw-rw-   1 root     root  %d Jan  1 1980 %s/%s\n",size,prefix,strbuf);
		    else printf("f %s\t(size %d)\n",strbuf,size);
		usb_control_msg(dh,IFP_REQ_TYPE,IFP_FILE_LEAVE,0,0,ctl,sizeof(ctl),TOUT); //??????
		break;
	    case 2:
		usb_bulk_read(dh,IFP_BULK_FROM,inbuf,sizeof(inbuf),TOUT);
		unicode2string(strbuf,inbuf);
		if (format)
		    printf("drw-rw-rw-   1 root     root  0 Jan  1 1980 %s/%s\n",prefix,strbuf);
		    else printf("d %s\n",strbuf);
		if (recurrent) {
		    strcpy(strnew,name);
		    if (strcmp(strnew,"/") != 0) strcat(strnew,"/");
		    strcat(strnew,strbuf);
		    push_dir(strnew);	// push directory with path for later use
		}
		break;
	    default:
		fprintf(stderr,"Uknown LS_NEXT return value = %d\n",ret);
		return -3;
	}
    }
    LS_END:
    usb_control_msg(dh,IFP_REQ_TYPE,IFP_LS_LEAVE,0,0,ctl,sizeof(ctl),TOUT);
    /* now go through the directories on the stack */
    while (pop_dir(strbuf)) {
	ifp_list_dir_printf(dh,strbuf,format,recurrent);
    }
    return 0;
}

int ifp_02_command(usb_dev_handle *dh, int number) {
    unsigned char ctl8[8];
    unsigned char bbuf[8];
    int i;
    usb_control_msg(dh,IFP_REQ_TYPE,IFP_02_COMMAND,0,number,ctl8,sizeof(ctl8),TOUT); // ?????
    //printf("0x02 ctl8: %d %d %d %d %d %d %d %d\n",ctl8[0],ctl8[1],ctl8[2],ctl8[3],ctl8[4],ctl8[5],ctl8[6],ctl8[7],ctl8[8]);
    if (!ifp_ok(ctl8)) return -1;
    i = usb_bulk_read(dh,IFP_BULK_FROM,bbuf,sizeof(bbuf),TOUT);
    //printf("Bulk (%d bytes): %d %d %d %d\n",i,bbuf[0],bbuf[1],bbuf[2],bbuf[3]);
    return 0;
}

int ifp_upload_file(usb_dev_handle *dh, const char *name, FILE *f) {
    int i,sum = 0;
    unsigned char buf[0x4000];
    unsigned char ctl[4];
    float per;
    int filesize = 0;	// this is a hack, original manager sends file size, this works
    struct stat stat;
    fstat(fileno(f),&stat);
    if (stat.st_size > ifp_get_free(dh)) {
	return -2;
    }

    rewind(f);
    ifp_send_filename(dh,name);
    usb_control_msg(dh,IFP_REQ_TYPE,IFP_FILE_UPLOAD,filesize,0,ctl,sizeof(ctl),TOUT);
    if (!ifp_ok(ctl)) return -1;
    while (!feof(f)) {
	i = fread(buf,1,sizeof(buf),f);
	//printf("%d bytes read form file...\n",i);
	printf("%d sent\r",sum);
	ifp_send_data(dh,buf,i);
	usb_control_msg(dh,IFP_REQ_TYPE,IFP_FILE_UPLOAD_APPEND,i,0,ctl,sizeof(ctl),TOUT);
	sum += i;
	//printf("upload-block-ctl: %d %d %d %d\n",ctl[0],ctl[1],ctl[2],ctl[3],ctl[4]);
    }
    printf("%d sent\r",sum);
    printf("\n");
    /* this is a hack - i've lost every last data block, so I send a dummy one */
    ifp_send_data(dh,buf,0);
    usb_control_msg(dh,IFP_REQ_TYPE,IFP_FILE_UPLOAD_APPEND,0,0,ctl,sizeof(ctl),TOUT);
    //printf("faked-upload-block-ctl: %d %d %d %d\n",ctl[0],ctl[1],ctl[2],ctl[3],ctl[4]);

    /* don't know why ? it work without too...
    ifp_02_command(dh,6);
    */

    usb_control_msg(dh,IFP_REQ_TYPE,IFP_FILE_LEAVE,0,0,ctl,sizeof(ctl),TOUT);
    if (!ifp_ok(ctl)) return -1;
    return 0;
}

int ifp_upload_dir(usb_dev_handle *dh, const char *from, const char *name) {
    FILE *fp;
    FTS *ftsp;
    FTSENT *ftsentp;
    size_t basename_pos;
    int i = -1;
    int fts_options = FTS_LOGICAL;
    char *basename = get_basename(from);
    char ifp_path[IFP_MAXPATHLEN];
    char *toname;
    char *argv[2] = {(char*)from, NULL};
    char *pos;

    basename_pos = strlen(from) - strlen(basename);

    if ( (ftsp = fts_open(argv, fts_options, comparefts)) == NULL) {
	perror("fts_open");
	goto FTS_END;
    }

    while ( (ftsentp = fts_read(ftsp)) != NULL) {

	toname = ftsentp->fts_path + (strlen(from) - strlen(basename));

	// "path" "file" -> "path/file"
	strncpy(ifp_path, name, IFP_MAXPATHLEN);
	strncat(ifp_path, "/", 1);
	strncat(ifp_path, toname, IFP_MAXPATHLEN - strlen(ifp_path) - 1);

	/* /path/to//file/// -> /path/to/file/ */
	remove_double_slash(ifp_path);
	/* /path/to/file/ -> /path/to/file */
	for (pos = ifp_path; *pos != '\0'; pos++);
	if (*(pos-1) == '/') {
	    *(pos-1) = '\0';
	}

	switch (ftsentp->fts_info) {
	    case FTS_F:		// normal file

		if ( (fp = fopen(ftsentp->fts_path, "r")) == NULL) {
		    perror("fopen");
		    goto FTS_END;
		}

		if ((i = ifp_upload_file(dh, ifp_path, fp)) < 0) {
		    fclose(fp);
		    goto FTS_END;
		}
		fclose(fp);

		break;

	    case FTS_D:		// directory

		i = ifp_make_dir(dh, ifp_path);
		break;

	    default:
		break;
	}
    }

FTS_END:
    if (fts_close(ftsp) == -1) {
	perror("fts_close");
	return -1;
    }
    return i;
}

int ifp_delete_file(usb_dev_handle *dh, const char *name) {
    unsigned char ctl[4];
    ifp_send_filename(dh,name);
    usb_control_msg(dh,IFP_REQ_TYPE,IFP_FILE_DELETE,0,0,ctl,sizeof(ctl),TOUT);
    if (!ifp_ok(ctl)) return -1;
    return 0;
}

// based ifp_list_dir_printf()
int ifp_delete_recursive(usb_dev_handle *dh, const char *name) {
    unsigned char ctl[4];
    unsigned char inbuf[IFP_BULK_MAXPATHLEN];
    unsigned char strbuf[IFP_MAXPATHLEN];
    unsigned char ifp_path[IFP_MAXPATHLEN];
    int i = -1;

    if ( (i = ifp_delete_file(dh, name)) == 0) {
	return i;
    }

    if (ifp_send_filename(dh, name) < 0)
	return -1;

    usb_control_msg(dh, IFP_REQ_TYPE, IFP_LS, 0, 0, ctl, sizeof(ctl), TOUT);
    if (!ifp_ok(ctl))
	return -1;

    while (1) {
	usb_control_msg(dh, IFP_REQ_TYPE, IFP_LS_NEXT, 3, 0, ctl, sizeof(ctl), TOUT);
	switch (ifp_ctl_ret(ctl)) {
	    case 0:	// end
		goto LS_END;
		break;
	    case 1:	// file
		usb_bulk_read(dh,IFP_BULK_FROM,inbuf,sizeof(inbuf),TOUT);
		unicode2string(strbuf,inbuf);

		// "path" "file" -> "path/file"
		strncpy(ifp_path, name, IFP_MAXPATHLEN);
		strncat(ifp_path, "/", 1);
		strncat(ifp_path, strbuf, IFP_MAXPATHLEN - strlen(ifp_path) - 1);
		i = ifp_delete_file(dh, ifp_path);

		break;
	    case 2:	// directory
		usb_bulk_read(dh,IFP_BULK_FROM,inbuf,sizeof(inbuf),TOUT);
		unicode2string(strbuf,inbuf);

		// "path" "file" -> "path/file"
		strncpy(ifp_path, name, IFP_MAXPATHLEN);
		strncat(ifp_path, "/", 1);
		strncat(ifp_path, strbuf, IFP_MAXPATHLEN - strlen(ifp_path) - 1);
		i = push_dir(ifp_path);

		break;
	    default:
		break;
	}
	if (i == -1) {
	    goto LS_END;
	}

    }

LS_END: 
    usb_control_msg(dh, IFP_REQ_TYPE, IFP_LS_LEAVE, 0, 0, ctl, sizeof(ctl), TOUT);
    while (pop_dir(strbuf)) {
	i = ifp_delete_recursive(dh, strbuf);
	if (i == -1) {
	    return -1;
	}
    }

    return ifp_delete_dir(dh, name);
}

int ifp_empty_dir(usb_dev_handle *dh, const char *name) {
    int i;
    unsigned char ctl[4];
    ifp_send_filename(dh,name);
    usb_control_msg(dh,IFP_REQ_TYPE,IFP_LS,0,0,ctl,sizeof(ctl),TOUT);
    if (!ifp_ok(ctl)) return -1;	// not a dir
    usb_control_msg(dh,IFP_REQ_TYPE,IFP_LS_NEXT,3,0,ctl,sizeof(ctl),TOUT);	// 3=magic !!!???
    i = (ifp_ctl_ret(ctl) == 0) ? 0 : 1;
    usb_control_msg(dh,IFP_REQ_TYPE,IFP_LS_LEAVE,0,0,ctl,sizeof(ctl),TOUT);
    return i;
}

int ifp_delete_dir(usb_dev_handle *dh, const char *name) {
    unsigned char ctl[4];
    ifp_send_filename(dh,name);
    usb_control_msg(dh,IFP_REQ_TYPE,IFP_DIR_DELETE,0,0,ctl,sizeof(ctl),TOUT);
    if (!ifp_ok(ctl)) return -1;
    return 0;
}

int ifp_make_dir(usb_dev_handle *dh, const char *name) {
    unsigned char ctl[4];
    ifp_send_filename(dh,name);
    usb_control_msg(dh,IFP_REQ_TYPE,IFP_DIR_CREATE,0,0,ctl,sizeof(ctl),TOUT);
    if (!ifp_ok(ctl)) return -1;
    return 0;
}

int ifp_download_file(usb_dev_handle *dh, const char *name, FILE *f) {
    int i,size,offset = 0;
    unsigned char ctl[4];
    unsigned char ctl8[8];
    unsigned char buf[0x4000];
    ifp_send_filename(dh,name);
    usb_control_msg(dh,IFP_REQ_TYPE,IFP_FILE_INFO,1,0,ctl,sizeof(ctl),TOUT);
    if (!ifp_ok(ctl)) return -1;
    usb_control_msg(dh,IFP_REQ_TYPE,IFP_GET_FILE_INFO_SIZE,0,0,ctl8,sizeof(ctl8),TOUT);
    do {
	usb_control_msg(dh,IFP_REQ_TYPE,IFP_FILE_DOWNLOAD,sizeof(buf),0,ctl,sizeof(ctl),TOUT);
	size = ctl[0] + ctl[1]*256;
	if (size > 0) {
	    i = usb_bulk_read(dh,IFP_BULK_FROM,buf,size,TOUT);
	    fwrite(buf,1,i,f);
	}
    } while (size > 0);
    usb_control_msg(dh,IFP_REQ_TYPE,IFP_FILE_LEAVE,0,0,ctl,sizeof(ctl),TOUT);
    if (!ifp_ok(ctl)) return -1;
    return 0;
}

/* not tested, not sure: */
int ifp_battery_condition(usb_dev_handle *dh) {
    unsigned char ctl8[8];
    unsigned char bbuf[8];
    int i;
    usb_control_msg(dh,IFP_REQ_TYPE,IFP_02_COMMAND,0,IFP_02_BATTERY,ctl8,sizeof(ctl8),TOUT);
    if (!ifp_ok(ctl8)) return -1;
    i = usb_bulk_read(dh,IFP_BULK_FROM,bbuf,sizeof(bbuf),TOUT);
    return bbuf[0];
}

int ifp_type_string(usb_dev_handle *dh, char *string) {
    unsigned char ctl8[8];
    unsigned char bbuf[32];
    int i;
    usb_control_msg(dh,IFP_REQ_TYPE,IFP_02_COMMAND,0,IFP_02_STRING,ctl8,sizeof(ctl8),TOUT);
    if (!ifp_ok(ctl8)) return -1;
    i = usb_bulk_read(dh,IFP_BULK_FROM,bbuf,sizeof(bbuf),TOUT);
    strcpy(string,bbuf);
    return 0;
}

int ifp_firmware_version(usb_dev_handle *dh) {
    unsigned char ctl8[8];
    unsigned char bbuf[8];
    int i;
    usb_control_msg(dh,IFP_REQ_TYPE,IFP_02_COMMAND,0,IFP_02_FIRMWARE,ctl8,sizeof(ctl8),TOUT);
    if (!ifp_ok(ctl8)) return -1;
    i = usb_bulk_read(dh,IFP_BULK_FROM,bbuf,sizeof(bbuf),TOUT);
    //printf("bulk: %d %d\n",bbuf[0],bbuf[1]);
    return ((bbuf[1]/16)*1000 + (bbuf[1]%16)*100 + (bbuf[0]/16)*10 + (bbuf[0]%16));
}

int ifp_check_connect(usb_dev_handle *dh) {
    return (ifp_battery_condition(dh) >= 0);
}

int ifp_format(usb_dev_handle *dh) {
    unsigned char ctl[4];
    usb_control_msg(dh, IFP_REQ_TYPE, IFP_FORMAT, 0, 0, ctl, sizeof(ctl), TOUT);
    if (!ifp_ok(ctl)) return -1;
    return 0;
}

int ifp_firmware_update(struct usb_device *dev, usb_dev_handle *dh, char *name) {
    FILE *fp;
    int i;
    unsigned char ctl[4];
    char *basename = get_basename(name);

    // send firmware file.
    if ( (strcmp(basename, "IFP-1XXTC.HEX") != 0) && (strcmp(basename, "IFP-1XXT.HEX") != 0) ) {
	fprintf(stderr, "Firmware filename must be set \"IFP-1XXTC.HEX\" or \"IFP-1XXT.HEX\".\n");
	return -1;
    }
    if ( (fp = fopen(name, "r")) == NULL) {
	perror("fopen");
	return -1;
    }

    if ( (i = ifp_upload_file(dh, basename, fp)) < 0) {
	fprintf(stderr, "Failed firmware upload.\n");
	fclose(fp);
	return i;
    }
    fclose(fp);
    printf("Please wait few minutes...\n");

    // send control message
    usb_control_msg(dh, IFP_REQ_TYPE, IFP_FIRMWARE_UPDATE, 0, 0, ctl, sizeof(ctl), TOUT);

    // close/reopen iFP for update.
    usb_close(dh);
    sleep(3);	// need this!

    if ( (dh = usb_open(dev)) == NULL) {
	fprintf(stderr, "unable to open iFP device\n");
	return -1;
    } else if (!ifp_check_connect(dh)) {
	fprintf(stderr,"unable to connect to iFP\n");
	return -1;
    }
    return i;
}

/****** support functions ************/

int push_dir(char *name) {
    dirlist *d;
    if ( (d = malloc(sizeof(dirlist))) == NULL) {
	return -1;
    }
    //printf("======================== Pushing dir %s\n",name);
    d->name = strdup(name);
    d->next = list;
    list = d;
    return 0;
}

int pop_dir(char *name) {
    dirlist *bak;
    if (list) {
	strcpy(name,list->name);
	//printf("========================= Poping dir %s\n",name);
	bak = list->next;
	free(list->name);
	free(list);
	list = bak;
	return 1;
    } else {
	return 0;
    }
}

int comparefts(const FTSENT **a, const FTSENT **b) {
    if ( (*a)->fts_info == FTS_ERR || (*b)->fts_info == FTS_ERR) {
	return 0;
    }
    if ( (*a)->fts_info == FTS_NS || (*b)->fts_info == FTS_NS) {
	if ( (*a)->fts_info != FTS_NS) {
	    return -1;
	} else if ( (*b)->fts_info != FTS_NS) {
	    return 1;
	} else {
	    return strcmp( (*a)->fts_name, (*b)->fts_name);
	}
    }
    return strcmp( (*a)->fts_name, (*b)->fts_name);


}

char *get_basename(const char *path) {
    char *start = (char*)path;

    for (; *path != '\0'; path++) {
	if (*path == '/') {
	    if (*(path+1) != '\0') {
		start = (char*)++path;
	    }
	}
    }

    return start;
}

void remove_double_slash(char *path) {
    char *pos = path;
    while (*path != '\0') {
	if (*path == '/' && *(path+1) == '/') {
	    path++;
	    continue;
	}
	*pos++ = *path++;
    }
    *pos = '\0';
}

